@@ -4,7 +4,7 @@
 ||| See the files COPYING and DISCLAIMER for more information.
 \*/
 #include "global.h"
-RCSID("$Id: las.c,v 1.43 1998/11/07 07:17:39 hubbe Exp $");
+RCSID("$Id: las.c,v 1.72 1999/02/04 20:33:56 hubbe Exp $");
 
 #include "language.h"
 #include "interpret.h"
@@ -26,6 +26,8 @@ RCSID("$Id: las.c,v 1.43 1998/11/07 07:17:39 hubbe Exp $");
 #include "callback.h"
 #include "pike_macros.h"
 #include "peep.h"
+#include "builtin_functions.h"
+#include "cyclic.h"
 
 #define LASDEBUG
 
@@ -34,7 +36,6 @@ int lasdebug=0;
 static node *eval(node *);
 static void optimize(node *n);
 
-dynamic_buffer areas[NUM_AREAS];
 node *init_node = 0;
 int num_parse_error;
 int cumulative_parse_error=0;
@@ -44,9 +45,9 @@ extern char *get_type_name(int);
 
 int car_is_node(node *n)
 {
-  check_node(n);
   switch(n->token)
   {
+  case F_EXTERNAL:
   case F_IDENTIFIER:
   case F_CONSTANT:
   case F_LOCAL:
@@ -59,9 +60,9 @@ int car_is_node(node *n)
 
 int cdr_is_node(node *n)
 {
-  check_node(n);
   switch(n->token)
   {
+  case F_EXTERNAL:
   case F_IDENTIFIER:
   case F_CONSTANT:
   case F_LOCAL:
@@ -73,11 +74,48 @@ int cdr_is_node(node *n)
   }
 }
 
+#ifdef PIKE_DEBUG
+void check_tree(node *n, int depth)
+{
+  if(!d_flag) return;
+  if(!n) return;
+  if(n->token==USHRT_MAX)
+    fatal("Free node in tree.\n");
+
+  if(d_flag<2) return;
+
+  if(!(depth & 1023))
+  {
+    node *q;
+    for(q=n->parent;q;q=q->parent)
+      if(q->parent==n)
+	fatal("Cyclic node structure found.\n");
+  }
+  depth++;
+
+  if(car_is_node(n))
+  {
+    if(CAR(n)->parent != n)
+      fatal("Parent is wrong.\n");
+
+    check_tree(CAR(n),depth);
+  }
+
+  if(cdr_is_node(n))
+  {
+    if(CDR(n)->parent != n)
+      fatal("Parent is wrong.\n");
+
+    check_tree(CDR(n),depth);
+  }
+}
+#endif
+
 INT32 count_args(node *n)
 {
   int a,b;
+  check_tree(n,0);
   if(!n) return 0;
-  check_node(n);
   switch(n->token)
   {
   case F_VAL_LVAL:
@@ -126,6 +164,32 @@ INT32 count_args(node *n)
   }
 }
 
+struct pike_string *find_return_type(node *n)
+{
+  struct pike_string *a,*b;
+
+  check_tree(n,0);
+
+  if(!n) return 0;
+  if(!(n->tree_info & OPT_RETURN)) return 0;
+  if(car_is_node(n))
+    a=find_return_type(CAR(n));
+  else
+    a=0;
+
+  if(cdr_is_node(n))
+    b=find_return_type(CDR(n));
+  else
+    b=0;
+
+  if(a)
+  {
+    if(b && a!=b) return mixed_type_string;
+    return a;
+  }
+  return b;
+}
+
 
 #define NODES 256
 
@@ -140,13 +204,13 @@ static node *free_nodes=0;
 
 void free_all_nodes(void)
 {
-  if(!local_variables)
+  if(!compiler_frame)
   {
     node *tmp;
     struct node_chunk *tmp2;
     int e=0;
 
-#ifndef DEBUG
+#ifndef PIKE_DEBUG
     if(cumulative_parse_error)
     {
 #endif
@@ -167,20 +231,18 @@ void free_all_nodes(void)
 	    if(!tmp)
 	    {
 	      tmp=tmp2->nodes+e;
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
 	      if(!cumulative_parse_error)
 	      {
-		fprintf(stderr,"Free node at %p, (%d) (token=%d).\n",tmp, tmp->line_number, tmp->token);
+		fprintf(stderr,"Free node at %p, (%s:%d) (token=%d).\n",tmp, tmp->current_file->str, tmp->line_number, tmp->token);
 		if(tmp->token==F_CONSTANT)
 		  print_tree(tmp);
 	      }
 	      else
 #endif
 	      {
-		check_node(tmp);
 		/* Free the node and be happy */
 		/* Make sure we don't free any nodes twice */
-		dmalloc_dump_track(tmp);
 		if(car_is_node(tmp)) CAR(tmp)=0;
 		if(cdr_is_node(tmp)) CDR(tmp)=0;
 		debug_malloc_touch(tmp->type);
@@ -189,12 +251,12 @@ void free_all_nodes(void)
 	    }
 	  }
 	}
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
 	if(!cumulative_parse_error)
 	  fatal("Failed to free %d nodes when compiling!\n",e2);
 #endif
       }
-#ifndef DEBUG
+#ifndef PIKE_DEBUG
     }
 #endif
     while(node_chunks)
@@ -208,10 +270,14 @@ void free_all_nodes(void)
   }
 }
 
-void debug_free_node(node *n)
+void free_node(node *n)
 {
   if(!n) return;
-  check_node(n);
+#ifdef PIKE_DEBUG
+  if(l_flag>9)
+    print_tree(n);
+#endif
+
   switch(n->token)
   {
   case USHRT_MAX:
@@ -228,7 +294,9 @@ void debug_free_node(node *n)
   }
   n->token=USHRT_MAX;
   if(n->type) free_string(n->type);
-  dmalloc_untrack(n);
+#ifdef PIKE_DEBUG
+  if(n->current_file) free_string(n->current_file);
+#endif
   CAR(n)=free_nodes;
   free_nodes=n;
 }
@@ -253,23 +321,40 @@ static node *mkemptynode(void)
   res=free_nodes;
   free_nodes=CAR(res);
   res->token=0;
-  res->line_number=current_line;
+  res->line_number=lex.current_line;
+#ifdef PIKE_DEBUG
+  copy_shared_string(res->current_file, lex.current_file);
+#endif
   res->type=0;
   res->node_info=0;
   res->tree_info=0;
   res->parent=0;
-  dmalloc_track(res);
   return res;
 }
 
-node *debug_mknode(short token,node *a,node *b)
+node *mknode(short token,node *a,node *b)
 {
   node *res;
+#ifdef PIKE_DEBUG
+  if(a && a==b)
+    fatal("mknode: a and be are the same!\n");
+#endif    
+    
+
+  check_tree(a,0);
+  check_tree(b,0);
   res = mkemptynode();
   CAR(res) = a;
   CDR(res) = b;
   res->node_info = 0;
   res->tree_info = 0;
+  if(a) a->parent = res;
+  if(b) b->parent = res;
+
+  /* We try to optimize most things, but argument lists are hard... */
+  if(token != F_ARG_LIST && (a || b))
+    res->node_info |= OPT_TRY_OPTIMIZE;
+
   switch(token)
   {
   case F_CATCH:
@@ -287,6 +372,10 @@ node *debug_mknode(short token,node *a,node *b)
     }
     break;
 
+  case F_UNDEFINED:
+    res->node_info |= OPT_EXTERNAL_DEPEND | OPT_SIDE_EFFECT;
+    break;
+
   case F_RETURN:
     res->node_info |= OPT_RETURN;
     break;
@@ -314,24 +403,34 @@ node *debug_mknode(short token,node *a,node *b)
   case F_POST_INC:
   case F_POST_DEC:
     res->node_info |= OPT_ASSIGNMENT;
+    break;
+    
+    default:
+      res->tree_info = res->node_info;
+      if(a) res->tree_info |= a->tree_info;
+      if(b) res->tree_info |= b->tree_info;
   }
   res->token = token;
   res->type = 0;
 
-  /* We try to optimize most things, but argument lists are hard... */
-  if(token != F_ARG_LIST && (a || b))
-    res->node_info |= OPT_TRY_OPTIMIZE;
 
-  if(a) a->parent = res;
-  if(b) b->parent = res;
-  if(!num_parse_error) optimize(res);
+#ifdef PIKE_DEBUG
+  if(d_flag > 3)
+    verify_shared_strings_tables();
+#endif
 
-#ifdef DEBUG
+  if(!num_parse_error && compiler_pass==2)
+  {
+    check_tree(res,0);
+    optimize(res);
+    check_tree(res,0);
+  }
+
+#ifdef PIKE_DEBUG
   if(d_flag > 3)
     verify_shared_strings_tables();
 #endif
 
-  check_node(res);
   return res;
 }
 
@@ -346,7 +445,6 @@ node *mkstrnode(struct pike_string *str)
   res->u.sval.subtype = 0;
 #endif
   copy_shared_string(res->u.sval.u.string, str);
-  check_node(res);
   return res;
 }
 
@@ -362,7 +460,6 @@ node *mkintnode(int nr)
   res->u.sval.type = T_INT;
   res->u.sval.subtype = NUMBER_NUMBER;
   res->u.sval.u.integer = nr;
-  check_node(res);
   return res;
 }
 
@@ -376,7 +473,6 @@ node *mkfloatnode(FLOAT_TYPE foo)
   res->u.sval.subtype = 0;
 #endif
   res->u.sval.u.float_number = foo;
-  check_node(res);
   return res;
 }
 
@@ -390,14 +486,12 @@ node *mkefuncallnode(char *function, node *args)
   struct pike_string *name;
   node *n;
   name = findstring(function);
-  if(!name || !find_module_identifier(name))
+  if(!name || !(n=find_module_identifier(name)))
   {
     my_yyerror("Internally used efun undefined: %s",function);
     return mkintnode(0);
   }
-  n=mkapplynode(mksvaluenode(sp-1), args);
-  pop_stack();
-  check_node(n);
+  n=mkapplynode(n, args);
   return n;
 }
 
@@ -413,14 +507,13 @@ node *mklocalnode(int var)
 {
   node *res = mkemptynode();
   res->token = F_LOCAL;
-  copy_shared_string(res->type, local_variables->variable[var].type);
+  copy_shared_string(res->type, compiler_frame->variable[var].type);
   res->node_info = OPT_NOT_CONST;
   res->tree_info=res->node_info;
 #ifdef __CHECKER__
   CDR(res)=0;
 #endif
   res->u.number = var;
-  check_node(res);
   return res;
 }
 
@@ -428,11 +521,10 @@ node *mkidentifiernode(int i)
 {
   node *res = mkemptynode();
   res->token = F_IDENTIFIER;
-  setup_fake_program();
-  copy_shared_string(res->type, ID_FROM_INT(&fake_program, i)->type);
+  copy_shared_string(res->type, ID_FROM_INT(new_program, i)->type);
 
   /* FIXME */
-  if(IDENTIFIER_IS_CONSTANT(ID_FROM_INT(&fake_program, i)->identifier_flags))
+  if(IDENTIFIER_IS_CONSTANT(ID_FROM_INT(new_program, i)->identifier_flags))
   {
     res->node_info = OPT_EXTERNAL_DEPEND;
   }else{
@@ -444,7 +536,33 @@ node *mkidentifiernode(int i)
   CDR(res)=0;
 #endif
   res->u.number = i;
-  check_node(res);
+  check_tree(res,0);
+  return res;
+}
+
+node *mkexternalnode(int level,
+		     int i,
+		     struct identifier *id)
+{
+  node *res = mkemptynode();
+  res->token = F_EXTERNAL;
+
+  copy_shared_string(res->type, id->type);
+
+  /* FIXME */
+  if(IDENTIFIER_IS_CONSTANT(ID_FROM_INT(parent_compilation(level), i)->identifier_flags))
+  {
+    res->node_info = OPT_EXTERNAL_DEPEND;
+  }else{
+    res->node_info = OPT_NOT_CONST;
+  }
+  res->tree_info=res->node_info;
+
+#ifdef __CHECKER__
+  CDR(res)=0;
+#endif
+  res->u.integer.a = level;
+  res->u.integer.b = i;
   return res;
 }
 
@@ -466,57 +584,122 @@ node *mkcastnode(struct pike_string *type,node *n)
   CDR(res)=0;
 #endif
   n->parent=res;
-  check_node(res);
   return res;
 }
 
 void resolv_constant(node *n)
 {
   struct identifier *i;
+  struct program *p;
+  INT32 numid;
+
+  check_tree(n,0);
 
   if(!n)
   {
     push_int(0);
   }else{
-    check_node(n);
     switch(n->token)
     {
     case F_CONSTANT:
       push_svalue(& n->u.sval);
+      return;
+
+    case F_EXTERNAL:
+      p=parent_compilation(n->u.integer.a);
+      if(!p)
+      {
+	yyerror("Failed to resolv external constant");
+	push_int(0);
+	return;
+      }
+      numid=n->u.integer.b;
       break;
 
     case F_IDENTIFIER:
-      setup_fake_program();
-      i=ID_FROM_INT(& fake_program, n->u.number);
-	
-      if(IDENTIFIER_IS_CONSTANT(i->identifier_flags))
-      {
-	push_svalue(PROG_FROM_INT(&fake_program, n->u.number)->constants +
-		    i->func.offset);
-      }else{
-	yyerror("Identifier is not a constant");
-	push_int(0);
-      }
+      p=new_program;
+      numid=n->u.number;
       break;
 
     case F_LOCAL:
-	yyerror("Expected constant, got local variable");
-	push_int(0);
-	break;
+      /* FIXME: Ought to have the name of the identifier in the message. */
+      yyerror("Expected constant, got local variable");
+      push_int(0);
+      return;
 
     case F_GLOBAL:
-	yyerror("Expected constant, got global variable");
-	push_int(0);
-	break;
+      /* FIXME: Ought to have the name of the identifier in the message. */
+      yyerror("Expected constant, got global variable");
+      push_int(0);
+      return;
+
+    case F_UNDEFINED:
+      if(compiler_pass==2) {
+	/* FIXME: Ought to have the name of the identifier in the message. */
+	yyerror("Expected constant, got undefined identifier");
+      }
+      push_int(0);
+      return;
+
+    default:
+      yyerror("Expected constant, got something else");
+      push_int(0);
+      return;
+    }
+
+    i=ID_FROM_INT(p, numid);
+    
+    if(IDENTIFIER_IS_CONSTANT(i->identifier_flags))
+    {
+      push_svalue(PROG_FROM_INT(p, numid)->constants +
+		  i->func.offset);
+    }else{
+      my_yyerror("Identifier '%s' is not a constant", i->name->str);
+      push_int(0);
     }
   }
 }
 
-node *debug_index_node(node *n, struct pike_string * id)
+void resolv_program(node *n)
+{
+  check_tree(n,0);
+
+  resolv_constant(n);
+  switch(sp[-1].type)
+  {
+    case T_OBJECT:
+      if(!sp[-1].u.object->prog)
+      {
+	pop_stack();
+	push_int(0);
+      }else{
+	f_object_program(1);
+      }
+      break;
+      
+    case T_FUNCTION:
+      if(program_from_function(sp-1))
+	break;
+      
+    default:
+      yyerror("Illegal program identifier");
+      pop_stack();
+      push_int(0);
+      
+    case T_PROGRAM:
+      break;
+  }
+}
+
+
+
+node *index_node(node *n, char *node_name, struct pike_string *id)
 {
   node *ret;
   JMP_BUF tmp;
-  check_node(n);
+
+  check_tree(n,0);
+
   if(SETJMP(tmp))
   {
     ONERROR tmp;
@@ -525,37 +708,69 @@ node *debug_index_node(node *n, struct pike_string * id)
     APPLY_MASTER("handle_error", 1);
     pop_stack();
     UNSET_ONERROR(tmp);
-    
-    yyerror("Couldn't index module.");
+
+    if (node_name) {
+      my_yyerror("Couldn't index module '%s'.", node_name);
+    } else {
+      yyerror("Couldn't index module.");
+    }
     push_int(0);
   }else{
     resolv_constant(n);
     switch(sp[-1].type)
     {
     case T_INT:
-      yyerror("Failed to index module (module doesn't exist?)");
+      if(!num_parse_error) {
+	if (node_name) {
+	  my_yyerror("Failed to index module '%s' (module doesn't exist?)",
+		     node_name);
+	} else {
+	  yyerror("Failed to index module (module doesn't exist?)");
+	}
+      }
       break;
 
-    case T_PROGRAM:
     case T_FLOAT:
     case T_STRING:
     case T_ARRAY:
-      yyerror("Failed to index module (Not a module?)");
+      if (node_name) {
+	my_yyerror("Failed to index module '%s' (Not a module?)",
+		   node_name);
+      } else {
+	yyerror("Failed to index module (Not a module?)");
+      }
       pop_stack();
       push_int(0);
       break;
 
     default:
-      ref_push_string(id);
-      f_index(2);
-
-      if(sp[-1].type == T_INT &&
-	 !sp[-1].u.integer &&
-	 sp[-1].subtype==NUMBER_UNDEFINED)
+    {
+      DECLARE_CYCLIC();
+      if(BEGIN_CYCLIC(sp[-1].u.refs, id))
       {
-	my_yyerror("Index '%s' not present in module.",id->str);
+	my_yyerror("Recursive module dependency in '%s'.",id->str);
+	pop_stack();
+	push_int(0);
+      }else{
+	SET_CYCLIC_RET(1);
+	ref_push_string(id);
+	f_index(2);
+      
+	if(sp[-1].type == T_INT &&
+	   !sp[-1].u.integer &&
+	   sp[-1].subtype==NUMBER_UNDEFINED)
+	{
+	  if (node_name) {
+	    my_yyerror("Index '%s' not present in module '%s'.",
+		       id->str, node_name);
+	  } else {
+	    my_yyerror("Index '%s' not present in module.", id->str);
+	  }
+	}
+	END_CYCLIC();
       }
     }
+    }
   }
   UNSETJMP(tmp);
   ret=mkconstantsvaluenode(sp-1);
@@ -566,6 +781,9 @@ node *debug_index_node(node *n, struct pike_string * id)
 
 int node_is_eq(node *a,node *b)
 {
+  check_tree(a,0);
+  check_tree(b,0);
+
   if(a == b) return 1;
   if(!a || !b) return 0;
   if(a->token != b->token) return 0;
@@ -627,12 +845,21 @@ node *mksvaluenode(struct svalue *s)
     return make_node_from_mapping(s->u.mapping);
 
   case T_OBJECT:
-    if(s->u.object == &fake_object)
+    if(s->u.object == fake_object)
     {
       return mkefuncallnode("this_object", 0);
-/*  }else{
-      yyerror("Non-constant object pointer! (should not happen!)");
- */
+    }
+    if(s->u.object->next == s->u.object)
+    {
+      int x=0;
+      struct object *o;
+      node *n=mkefuncallnode("this_object", 0);
+      for(o=fake_object;o!=s->u.object;o=o->parent)
+      {
+	n=mkefuncallnode("function_object",
+			 mkefuncallnode("object_program",n));
+      }
+      return n;
     }
     break;
 
@@ -640,9 +867,19 @@ node *mksvaluenode(struct svalue *s)
   {
     if(s->subtype != FUNCTION_BUILTIN)
     {
-      if(s->u.object == &fake_object)
+      if(s->u.object == fake_object)
 	return mkidentifiernode(s->subtype);
 
+      if(s->u.object->next == s->u.object)
+      {
+	int x=0;
+	struct object *o;
+	for(o=fake_object->parent;o!=s->u.object;o=o->parent) x++;
+	return mkexternalnode(x, s->subtype,
+			      ID_FROM_INT(o->prog, s->subtype));
+
+      }
+
 /*      yyerror("Non-constant function pointer! (should not happen!)"); */
     }
   }
@@ -659,6 +896,7 @@ node *mksvaluenode(struct svalue *s)
 node *copy_node(node *n)
 {
   node *b;
+  check_tree(n,0);
   if(!n) return n;
   switch(n->token)
   {
@@ -754,6 +992,13 @@ int node_is_false(node *n)
   }
 }
 
+int node_may_overload(node *n, int lfun)
+{
+  if(!n) return 0;
+  if(!n->type) return 1;
+  return type_may_overload(n->type->str, lfun);
+}
+
 node **last_cmd(node **a)
 {
   node **n;
@@ -799,11 +1044,30 @@ static node **low_get_arg(node **a,int *nr)
 }
 
 node **my_get_arg(node **a,int n) { return low_get_arg(a,&n); }
-/* static node **first_arg(node **a) { return my_get_arg(a,0); } */
+
+node **is_call_to(node *n, c_fun f)
+{
+  switch(n->token)
+  {
+    case F_APPLY:
+      if(CAR(n) &&
+	 CAR(n)->token == F_CONSTANT &&
+	 CAR(n)->u.sval.type == T_FUNCTION &&
+	 CAR(n)->u.sval.subtype == FUNCTION_BUILTIN &&
+	 CAR(n)->u.sval.u.efun->function == f)
+	return & CDR(n);
+  }
+  return 0;
+}
+
 
 static void low_print_tree(node *foo,int needlval)
 {
   if(!foo) return;
+  if(l_flag>9)
+  {
+    printf("/*%x*/",foo->tree_info);
+  }
   switch(foo->token)
   {
   case F_LOCAL:
@@ -823,8 +1087,7 @@ static void low_print_tree(node *foo,int needlval)
 
   case F_IDENTIFIER:
     if(needlval) putchar('&');
-    setup_fake_program();
-    printf("%s",ID_FROM_INT(&fake_program, foo->u.number)->name->str);
+    printf("%s",ID_FROM_INT(new_program, foo->u.number)->name->str);
     break;
 
   case F_ASSIGN:
@@ -859,6 +1122,12 @@ static void low_print_tree(node *foo,int needlval)
     low_print_tree(CDR(foo),needlval);
     return;
 
+    case F_ARRAY_LVALUE:
+      putchar('[');
+      low_print_tree(CAR(foo),1);
+      putchar(']');
+      break;
+
   case F_LVALUE_LIST:
     low_print_tree(CAR(foo),1);
     if(CAR(foo) && CDR(foo)) putchar(',');
@@ -912,6 +1181,7 @@ static void low_print_tree(node *foo,int needlval)
 
 void print_tree(node *n)
 {
+  check_tree(n,0);
   low_print_tree(n,0);
   printf("\n");
   fflush(stdout);
@@ -1040,7 +1310,7 @@ static void find_written_vars(node *n,
     if(lvalue) p->locals[n->u.number]=VAR_USED;
     break;
 
-  case F_GLOBAL:
+  case F_IDENTIFIER:
      if(lvalue)
      {
        if(n->u.number>=MAX_GLOBAL)
@@ -1094,6 +1364,10 @@ static void find_written_vars(node *n,
     find_written_vars(CDR(n), p, 1);
     break;
 
+    case F_ARRAY_LVALUE:
+    find_written_vars(CAR(n), p, 1);
+    break;
+      
   case F_LVALUE_LIST:
     find_written_vars(CAR(n), p, 1);
     find_written_vars(CDR(n), p, 1);
@@ -1139,7 +1413,7 @@ static int depend_p2(node *a,node *b)
   return 0;
 }
 
-static int depend_p(node *a,node *b)
+static int depend_p3(node *a,node *b)
 {
   if(!b) return 0;
 #if 0
@@ -1153,6 +1427,31 @@ static int depend_p(node *a,node *b)
   return depend_p2(a,b);
 }
 
+#ifdef PIKE_DEBUG
+static int depend_p(node *a,node *b)
+{
+  int ret;
+  if(l_flag > 3)
+  {
+    fprintf(stderr,"Checking if: ");
+    print_tree(a);
+    fprintf(stderr,"Depends on: ");
+    print_tree(b);
+    if(depend_p3(a,b))
+    {
+      fprintf(stderr,"The answer is (durumroll) : yes\n");
+      return 1;
+    }else{
+      fprintf(stderr,"The answer is (durumroll) : no\n");
+      return 0;
+    }
+  }
+  return depend_p3(a,b);
+}
+#else
+#define depend_p depend_p3
+#endif
+
 static int cntargs(node *n)
 {
   if(!n) return 0;
@@ -1211,8 +1510,6 @@ void fix_type_field(node *n)
 {
   struct pike_string *type_a,*type_b;
 
-  check_node(n);
-
   if(n->type) return; /* assume it is correct */
 
   switch(n->token)
@@ -1237,31 +1534,24 @@ void fix_type_field(node *n)
     if(CAR(n) && CDR(n) && 
        !match_types(CDR(n)->type,CAR(n)->type))
       my_yyerror("Bad type in assignment.");
-    copy_shared_string(n->type, CDR(n)->type);
+    copy_shared_string(n->type, CAR(n)->type);
     break;
 
   case F_INDEX:
-    type_a=CAR(n)->type;
-    type_b=CDR(n)->type;
-    if(!check_indexing(type_a, type_b, n))
-      my_yyerror("Indexing on illegal type.");
-    n->type=index_type(type_a,n);
-    check_node(n);
-    break;
-
   case F_ARROW:
     type_a=CAR(n)->type;
     type_b=CDR(n)->type;
     if(!check_indexing(type_a, type_b, n))
-      my_yyerror("Indexing on illegal type.");
+      if(!catch_level)
+        my_yyerror("Indexing on illegal type.");
     n->type=index_type(type_a,n);
-    check_node(n);
     break;
 
   case F_APPLY:
   {
     struct pike_string *s;
     struct pike_string *f;
+    INT32 max_args,args;
     push_type(T_MIXED); /* match any return type, even void */
     push_type(T_VOID); /* not varargs */
     push_type(T_MANY);
@@ -1270,31 +1560,61 @@ void fix_type_field(node *n)
     s=pop_type();
     f=CAR(n)->type?CAR(n)->type:mixed_type_string;
     n->type=check_call(s,f);
+    args=count_arguments(s);
+    max_args=count_arguments(f);
+    if(max_args<0) max_args=0x7fffffff;
 
     if(!n->type)
     {
       char *name;
-      int args;
       switch(CAR(n)->token)
       {
       case F_IDENTIFIER:
-	setup_fake_program();
-	name=ID_FROM_INT(&fake_program, CAR(n)->u.number)->name->str;
+	name=ID_FROM_INT(new_program, CAR(n)->u.number)->name->str;
 	break;
 	
       case F_CONSTANT:
+	switch(CAR(n)->u.sval.type)
+	{
+	  case T_FUNCTION:
+	    if(CAR(n)->u.sval.subtype == FUNCTION_BUILTIN)
+	    {
+	      name=CAR(n)->u.sval.u.efun->name->str;
+	    }else{
+	      name=ID_FROM_INT(CAR(n)->u.sval.u.object->prog,
+			       CAR(n)->u.sval.subtype)->name->str;
+	    }
+	    break;
+
+	  case T_ARRAY:
+	    name="array call";
+	    break;
+
+	  case T_PROGRAM:
+	    name="clone call";
+	    break;
+
+	  default:
+	    name="`() (function call)";
+	    break;
+	}
+	break;
+
       default:
-	name="`() (function call)";
+	name="unknown function";
       }
 
-      if(max_correct_args == count_arguments(s))
+      if(max_args < args)
       {
-	my_yyerror("To few arguments to %s.\n",name);
+	my_yyerror("Too many arguments to %s.\n",name);
+      }
+      else if(max_correct_args == args)
+      {
+	my_yyerror("Too few arguments to %s.\n",name);
       }else{
 	my_yyerror("Bad argument %d to %s.",
 		   max_correct_args+1, name);
       }
-      check_node(n);
       copy_shared_string(n->type, mixed_type_string);
     }
     free_string(s);
@@ -1325,11 +1645,11 @@ void fix_type_field(node *n)
     {
       yyerror("You cannot return a void expression");
     }
-    if(local_variables &&
-       local_variables->current_return_type &&
-       !match_types(local_variables->current_return_type,CAR(n)->type) &&
+    if(compiler_frame &&
+       compiler_frame->current_return_type &&
+       !match_types(compiler_frame->current_return_type,CAR(n)->type) &&
        !(
-	 local_variables->current_return_type==void_type_string &&
+	 compiler_frame->current_return_type==void_type_string &&
 	 CAR(n)->token == F_CONSTANT &&
 	 IS_ZERO(& CAR(n)->u.sval)
 	 )
@@ -1394,13 +1714,11 @@ void fix_type_field(node *n)
   default:
     copy_shared_string(n->type,mixed_type_string);
   }
-  check_node(n);
 }
 
 static void zapp_try_optimize(node *n)
 {
   if(!n) return;
-  check_node(n);
   n->node_info &=~ OPT_TRY_OPTIMIZE;
   n->tree_info &=~ OPT_TRY_OPTIMIZE;
   if(car_is_node(n)) zapp_try_optimize(CAR(n));
@@ -1410,8 +1728,7 @@ static void zapp_try_optimize(node *n)
 static void optimize(node *n)
 {
   node *tmp1, *tmp2, *tmp3;
-  INT32 save_line = current_line;
-  check_node(n);
+  INT32 save_line = lex.current_line;
   do
   {
     if(car_is_node(n) && !(CAR(n)->node_info & OPT_OPTIMIZED))
@@ -1424,7 +1741,7 @@ static void optimize(node *n)
       n=CDR(n);
       continue;
     }
-    current_line = n->line_number;
+    lex.current_line = n->line_number;
 
 
     n->tree_info = n->node_info;
@@ -1466,8 +1783,9 @@ static void optimize(node *n)
       }
     }
     fix_type_field(n);
+    debug_malloc_touch(n->type);
 
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
     if(l_flag > 3 && n)
     {
       fprintf(stderr,"Optimizing (tree info=%x):",n->tree_info);
@@ -1475,8 +1793,6 @@ static void optimize(node *n)
     }
 #endif    
 
-    check_node(n);
-
     switch(n->token)
     {
     case F_APPLY:
@@ -1551,21 +1867,21 @@ static void optimize(node *n)
 	if(CDR(n)->token == F_CONSTANT && CDR(n)->u.sval.type == T_INT)
 	{
 	  /* a+=0 -> a */
-	  if(n->u.sval.u.integer == 0) goto use_car;
+	  if(CDR(n)->u.sval.u.integer == 0) goto use_car;
 
 	  /* a+=1 -> ++a */
-	  if(n->u.sval.u.integer == 1)
+	  if(CDR(n)->u.sval.u.integer == 1)
 	  {
-	    tmp1=mknode(F_INC,CDR(n),0);
-	    CDR(n)=0;
+	    tmp1=mknode(F_INC,CAR(n),0);
+	    CAR(n)=0;
 	    goto use_tmp1;
 	  }
 
 	  /* a+=-1 -> --a */
-	  if(n->u.sval.u.integer == -1)
+	  if(CDR(n)->u.sval.u.integer == -1)
 	  {
-	    tmp1=mknode(F_DEC, CDR(n), 0);
-	    CDR(n)=0;
+	    tmp1=mknode(F_DEC, CAR(n), 0);
+	    CAR(n)=0;
 	    goto use_tmp1;
 	  }
 	}
@@ -1578,21 +1894,21 @@ static void optimize(node *n)
 	if(CDR(n)->token == F_CONSTANT && CDR(n)->u.sval.type == T_INT)
 	{
 	  /* a-=0 -> a */
-	  if(n->u.sval.u.integer == 0) goto use_car;
+	  if(CDR(n)->u.sval.u.integer == 0) goto use_car;
 
 	  /* a-=-1 -> ++a */
-	  if(n->u.sval.u.integer == -1)
+	  if(CDR(n)->u.sval.u.integer == -1)
 	  {
-	    tmp1=mknode(F_INC, CDR(n), 0);
-	    CDR(n)=0;
+	    tmp1=mknode(F_INC, CAR(n), 0);
+	    CAR(n)=0;
 	    goto use_tmp1;
 	  }
 
 	  /* a-=-1 -> --a */
-	  if(n->u.sval.u.integer == 1)
+	  if(CDR(n)->u.sval.u.integer == 1)
 	  {
-	    tmp1=mknode(F_DEC, CDR(n), 0);
-	    CDR(n)=0;
+	    tmp1=mknode(F_DEC, CAR(n), 0);
+	    CAR(n)=0;
 	    goto use_tmp1;
 	  }
 	}
@@ -1750,14 +2066,15 @@ static void optimize(node *n)
 	if(!node_is_eq(*arg1,CAR(tmp1)) || /* x == z */
 	   depend_p(*arg2,*arg2) ||	/* does y depend on y? */
 	   depend_p(*arg2,*arg1) ||	/* does y depend on x? */
-	   depend_p(*arg2,CADR(n))) /* does y depend on p? */
+	   depend_p(*arg2,CADR(n)) ||   /* does y depend on p? */
+	   depend_p(*arg2,tmp1))        /* does y depend on z? */
 	{
 	  /* it was not on the form for(; x op y; x++) p; */
 	  if(!node_is_eq(*arg2,CAR(tmp1)) || /* y == z */
 	     depend_p(*arg1,*arg2) || /* does x depend on y? */
 	     depend_p(*arg1,*arg1) || /* does x depend on x? */
-	     depend_p(*arg1,CADR(n)) /* does x depend on p? */
-	     )
+	     depend_p(*arg1,CADR(n)) || /* does x depend on p? */
+	     depend_p(*arg1,tmp1))      /* does x depend on z? */
 	  {
 	    /* it was not on the form for(; x op y; y++) p; */
 	    break;
@@ -1843,7 +2160,7 @@ static void optimize(node *n)
       if(tmp1) tmp1->parent = n->parent;
       free_node(n);
       n=tmp1;
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
       if(l_flag > 3)
       {
 	fprintf(stderr,"Result: ");
@@ -1856,7 +2173,7 @@ static void optimize(node *n)
     n->node_info |= OPT_OPTIMIZED;
     n=n->parent;
   }while(n);
-  current_line = save_line;
+  lex.current_line = save_line;
 }
 
 struct timer_oflo
@@ -1871,7 +2188,7 @@ static void check_evaluation_time(struct callback *cb,void *tmp,void *ignored)
   if(foo->counter-- < 0)
   {
     foo->yes=1;
-    throw();
+    pike_throw();
   }
 }
 
@@ -1882,7 +2199,7 @@ int eval_low(node *n)
   struct svalue *save_sp = sp;
   int ret;
 
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
   if(l_flag > 3 && n)
   {
     fprintf(stderr,"Evaluating (tree info=%x):",n->tree_info);
@@ -1890,11 +2207,10 @@ int eval_low(node *n)
   }
 #endif
 
-  if(num_parse_error) return -1;
-  setup_fake_program();
+  if(num_parse_error) return -1; 
 
-  num_strings=fake_program.num_strings;
-  num_constants=fake_program.num_constants;
+  num_strings=new_program->num_strings;
+  num_constants=new_program->num_constants;
   jump=PC;
 
   store_linenumbers=0;
@@ -1902,7 +2218,6 @@ int eval_low(node *n)
   ins_f_byte(F_DUMB_RETURN);
   store_linenumbers=1;
 
-  setup_fake_program();
   ret=-1;
   if(!num_parse_error)
   {
@@ -1913,27 +2228,28 @@ int eval_low(node *n)
     foo.counter=10000;
     foo.yes=0;
 
-    setup_fake_object();
-
     tmp_callback=add_to_callback(&evaluator_callbacks,
 				 check_evaluation_time,
 				 (void *)&foo,0);
 				 
-    if(apply_low_safe_and_stupid(&fake_object, jump))
+    if(apply_low_safe_and_stupid(fake_object, jump))
     {
       /* Generate error message */
-      if(throw_value.type == T_ARRAY && throw_value.u.array->size)
+      if(!catch_level)
       {
-	union anything *a;
-	a=low_array_get_item_ptr(throw_value.u.array, 0, T_STRING);
-	if(a)
-	{
-	  yyerror(a->string->str);
+        if(throw_value.type == T_ARRAY && throw_value.u.array->size)
+        {
+	  union anything *a;
+	  a=low_array_get_item_ptr(throw_value.u.array, 0, T_STRING);
+	  if(a)
+	  {
+	    yyerror(a->string->str);
+	  }else{
+	    yyerror("Nonstandard error format.");
+	  }
 	}else{
 	  yyerror("Nonstandard error format.");
 	}
-      }else{
-	yyerror("Nonstandard error format.");
       }
     }else{
       if(foo.yes)
@@ -1945,21 +2261,19 @@ int eval_low(node *n)
     remove_callback(tmp_callback);
   }
 
-  while(fake_program.num_strings > num_strings)
+  while(new_program->num_strings > num_strings)
   {
-    fake_program.num_strings--;
-    free_string(fake_program.strings[fake_program.num_strings]);
-    areas[A_STRINGS].s.len-=sizeof(struct pike_string *);
+    new_program->num_strings--;
+    free_string(new_program->strings[new_program->num_strings]);
   }
 
-  while(fake_program.num_constants > num_constants)
+  while(new_program->num_constants > num_constants)
   {
-    fake_program.num_constants--;
-    free_svalue(fake_program.constants + fake_program.num_constants);
-    areas[A_CONSTANTS].s.len-=sizeof(struct svalue);
+    new_program->num_constants--;
+    free_svalue(new_program->constants + new_program->num_constants);
   }
 
-  areas[A_PROGRAM].s.len=jump;
+  new_program->num_program=jump;
 
   return ret;
 }
@@ -1979,11 +2293,17 @@ static node *eval(node *n)
     break;
 
   case 0:
+    if(catch_level) return n;
     free_node(n);
     n=0;
     break;
 
   case 1:
+    if(catch_level && IS_ZERO(sp-1))
+    {
+      pop_stack();
+      return n;
+    }
     free_node(n);
     n=mksvaluenode(sp-1);
     pop_stack();
@@ -2073,7 +2393,11 @@ static struct svalue *is_stupid_func(node *n,
 
   n=CAR(n);
   if(!n || n->token != F_CONSTANT) return 0;
-  return &n->u.sval;
+
+  if((count_arguments(n->type) < 0) == !!vargs)
+    return &n->u.sval;
+
+  return 0;
 }
 
 int dooptcode(struct pike_string *name,
@@ -2085,16 +2409,12 @@ int dooptcode(struct pike_string *name,
   int args, vargs, ret;
   struct svalue *foo;
 
-#ifdef DEBUG
-  if(recoveries && sp-evaluator_stack < recoveries->sp)
-    fatal("Stack error before dooptcode (underflow)\n");
-#endif
+  check_tree(n,0);
 
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
   if(a_flag > 1)
     fprintf(stderr,"Doing function '%s' at %x\n",name->str,PC);
 #endif
-  last_function_opt_info=OPT_SIDE_EFFECT;
 
   args=count_arguments(type);
   if(args < 0) 
@@ -2104,54 +2424,73 @@ int dooptcode(struct pike_string *name,
   }else{
     vargs=0;
   }
-  n=mknode(F_ARG_LIST,n,0);
-
-  if((foo=is_stupid_func(n, args, vargs)))
+  if(compiler_pass==1)
   {
-    if(foo->type == T_FUNCTION && foo->subtype==FUNCTION_BUILTIN)
+    tmp.offset=-1;
+#ifdef PIKE_DEBUG
+    if(a_flag > 4)
     {
-      tmp.c_fun=foo->u.efun->function;
-      ret=define_function(name,
-			  type,
-			  modifiers,
-			  IDENTIFIER_C_FUNCTION | vargs,
-			  &tmp);
-      free_node(n);
-      return ret;
+      fprintf(stderr,"Making prototype (pass 1) for: ");
+      print_tree(n);
+    }
+#endif
+  }else{
+    n=mknode(F_ARG_LIST,n,0);
+    
+    if((foo=is_stupid_func(n, args, vargs)))
+    {
+      if(foo->type == T_FUNCTION && foo->subtype==FUNCTION_BUILTIN)
+      {
+	tmp.c_fun=foo->u.efun->function;
+	ret=define_function(name,
+			    type,
+			    modifiers,
+			    IDENTIFIER_C_FUNCTION | vargs,
+			    &tmp);
+	free_node(n);
+#ifdef PIKE_DEBUG
+	if(a_flag > 1)
+	  fprintf(stderr,"Identifer (C) = %d\n",ret);
+#endif
+	return ret;
+      }
+    }
+
+    tmp.offset=PC;
+    add_to_program(compiler_frame->max_number_of_locals);
+    add_to_program(args);
+  
+#ifdef PIKE_DEBUG
+    if(a_flag > 2)
+    {
+      fprintf(stderr,"Coding: ");
+      print_tree(n);
+    }
+#endif
+    if(!num_parse_error)
+    {
+      extern int remove_clear_locals;
+      remove_clear_locals=args;
+      if(vargs) remove_clear_locals++;
+      do_code_block(n);
+      remove_clear_locals=0x7fffffff;
     }
   }
   
-  tmp.offset=PC;
-  ins_byte(local_variables->max_number_of_locals, A_PROGRAM);
-  ins_byte(args, A_PROGRAM);
-  
-#ifdef DEBUG
-  if(a_flag > 2)
-  {
-    fprintf(stderr,"Coding: ");
-    print_tree(n);
-  }
-#endif
-  if(!num_parse_error)
-  {
-    do_code_block(n);
-  }
-  
-#ifdef DEBUG
-  if(recoveries && sp-evaluator_stack < recoveries->sp)
-    fatal("Stack error after do_code_block (underflow)\n");
-#endif
-
   ret=define_function(name,
 		      type,
 		      modifiers,
 		      IDENTIFIER_PIKE_FUNCTION | vargs,
 		      &tmp);
 
+
+#ifdef PIKE_DEBUG
+  if(a_flag > 1)
+    fprintf(stderr,"Identifer = %d\n",ret);
+#endif
+
   free_node(n);
   return ret;
 }
 
-INT32 get_opt_info(void) { return last_function_opt_info; }
-
 
