@@ -5,6 +5,10 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 
+#ifdef HAVE_SYS_FILE_H
+#include <sys/file.h>
+#endif /* HAVE_SYS_FILE_H */
+
 #include <errno.h>
 
 #ifdef HAVE_SYS_MMAN_H
@@ -22,7 +26,7 @@
 
 #include <fcntl.h>
 
-RCSID("$Id: pipe.c,v 1.19 1998/06/22 16:39:33 grubba Exp $");
+RCSID("$Id: pipe.c,v 1.29 1999/07/29 16:48:03 grubba Exp $");
 
 #include "threads.h"
 #include "stralloc.h"
@@ -49,6 +53,29 @@ RCSID("$Id: pipe.c,v 1.19 1998/06/22 16:39:33 grubba Exp $");
 #define BLOCKING_CLOSE
 */
 
+#ifndef SEEK_SET
+#ifdef L_SET
+#define SEEK_SET	L_SET
+#else /* !L_SET */
+#define SEEK_SET	0
+#endif /* L_SET */
+#endif /* SEEK_SET */
+#ifndef SEEK_CUR
+#ifdef L_INCR
+#define SEEK_CUR	L_INCR
+#else /* !L_INCR */
+#define SEEK_CUR	1
+#endif /* L_INCR */
+#endif /* SEEK_CUR */
+#ifndef SEEK_END
+#ifdef L_XTND
+#define SEEK_END	L_XTND
+#else /* !L_XTND */
+#define SEEK_END	2
+#endif /* L_XTND */
+#endif /* SEEK_END */
+
+
 #if 0
 #define INSISTANT_WRITE
 #endif
@@ -71,6 +98,9 @@ RCSID("$Id: pipe.c,v 1.19 1998/06/22 16:39:33 grubba Exp $");
 
 static struct program *pipe_program, *output_program;
 
+#ifdef THIS
+#undef THIS
+#endif
 #define THIS ((struct pipe *)(fp->current_storage))
 #define THISOBJ (fp->current_object)
 
@@ -155,8 +185,7 @@ static INLINE void output_try_write_some(struct object *obj);
  */
 static void push_callback(int no)
 {
-  sp->u.object=THISOBJ;
-  THISOBJ->refs++;
+  add_ref(sp->u.object=THISOBJ);
   sp->subtype=no+fp->context.identifier_level;
   sp->type=T_FUNCTION;
   sp++;
@@ -263,8 +292,8 @@ static INLINE int append_buffer(struct pike_string *s)
 
    if(THIS->fd!= -1)
    {
-     lseek(THIS->fd,THIS->pos,0);
-     write(THIS->fd,s->str,s->len);
+     lseek(THIS->fd, THIS->pos, SEEK_SET);
+     write(THIS->fd, s->str, s->len);
      THIS->pos+=s->len;
      return 0;
    }
@@ -275,7 +304,7 @@ static INLINE int append_buffer(struct pike_string *s)
      b->next=NULL;
      b->s=s;
      sbuffers += s->len;
-     s->refs++;
+     add_ref(s);
 
      if (THIS->lastbuffer)
        THIS->lastbuffer->next=b;
@@ -296,10 +325,10 @@ static void low_start(void)
   struct output *o;
 
 
-  THISOBJ->refs++;		/* dont kill yourself now */
+  add_ref(THISOBJ);		/* dont kill yourself now */
   for(obj=THIS->firstoutput;obj;obj=next)
   {
-    obj->refs++; /* Hang on PLEASE!! /hubbe */
+    add_ref(obj);		/* Hang on PLEASE!! /hubbe */
     o=(struct output *)(obj->storage);
     if (o->obj && o->mode==O_SLEEP)
     {
@@ -326,6 +355,9 @@ static void low_start(void)
   free_object(THISOBJ);
 }
 
+/* Read some data from the blocking object.
+ *
+ */
 static int read_some_data(void)
 {
   struct pipe *this = THIS;
@@ -333,7 +365,7 @@ static int read_some_data(void)
 
   if (!i || i->type != I_BLOCKING_OBJ) {
     fatal("PIPE: read_some_data(): Bad input type!\n");
-    return(0);	/* Keep compiler happy */
+    return -1;
   }
   push_int(8192);
   push_int(1);    /* We don't care if we don't get all 8192 bytes. */
@@ -342,13 +374,13 @@ static int read_some_data(void)
     append_buffer(sp[-1].u.string);
     pop_stack();
     THIS->sleeping = 1;
-    return(1);	/* Success */
+    return(1);        /* Success */
   }
 
   /* FIXME: Should we check the return value here? */
   pop_stack();
   /* EOF */
-  return(0);	/* EOF */
+  return(0);  /* EOF */
 }
 
 /* Let's guess what this function does....
@@ -417,7 +449,7 @@ static INLINE struct pike_string* gimme_some_data(unsigned long pos)
       len=this->pos-pos;
       if (len>READ_BUFFER_SIZE) len=READ_BUFFER_SIZE;
       THREADS_ALLOW();
-      lseek(this->fd, pos, 0); /* SEEK_SET */
+      lseek(this->fd, pos, SEEK_SET);
       THREADS_DISALLOW();
       do {
 	THREADS_ALLOW();
@@ -429,6 +461,9 @@ static INLINE struct pike_string* gimme_some_data(unsigned long pos)
 	  }
 	}
       } while(len < 0);
+      /*
+       * FIXME: What if len is 0?
+       */
       return make_shared_binary_string(buffer,len);
    }
 
@@ -512,7 +547,7 @@ static INLINE struct pike_string* gimme_some_data(unsigned long pos)
 
    if (pos==this->pos)
    {
-      this->firstbuffer->s->refs++;
+      add_ref(this->firstbuffer->s);
       return this->firstbuffer->s;
    }
    return make_shared_binary_string(this->firstbuffer->s->str+
@@ -668,7 +703,7 @@ static void pipe_input(INT32 args)
 	 i->type=I_MMAP;
 	 i->len=s.st_size;
 	 i->u.mmap=m;
-#ifdef HAVE_MADVISE
+#if defined(HAVE_MADVISE) && defined(MADV_SEQUENTIAL)
 	 /* Mark the pages as sequential read only access... */
 
 	 /* NOTE:
@@ -681,7 +716,7 @@ static void pipe_input(INT32 args)
 #ifdef HAVE_SETEUID
 	   seteuid(0);
 #else /* ! HAVE_SETEUID */
-#ifdef HAVE_SETREUID
+#ifdef HAVE_SETRESUID
 	   setresuid(-1, 0, -1);
 #endif /* HAVE_SETRESUID */
 #endif /* HAVE_SETEUID */
@@ -693,7 +728,7 @@ static void pipe_input(INT32 args)
 #ifdef HAVE_SETEUID
 	   seteuid(ou);
 #else /* ! HAVE_SETEUID */
-#ifdef HAVE_SETREUID
+#ifdef HAVE_SETRESUID
 	   setresuid(-1, ou, -1);
 #endif /* HAVE_SETRESUID */
 #endif /* HAVE_SETEUID */
@@ -711,7 +746,7 @@ static void pipe_input(INT32 args)
    i->u.obj=obj;
    nobjects++;
    i->type=I_OBJ;
-   i->u.obj->refs++;
+   add_ref(i->u.obj);
    i->set_nonblocking_offset=find_identifier("set_nonblocking",i->u.obj->prog);
    i->set_blocking_offset=find_identifier("set_blocking",i->u.obj->prog);
 
@@ -719,22 +754,25 @@ static void pipe_input(INT32 args)
        i->set_blocking_offset<0) 
    {
       if (find_identifier("read", i->u.obj->prog) < 0) {
-	/* Not even a read function */
-	free_object(i->u.obj);
-	i->u.obj=NULL;
-	i->type=I_NONE;
+	 /* Not even a read function */
+	 free_object(i->u.obj);
+	 i->u.obj=NULL;
+	 i->type=I_NONE;
 
-	nobjects--;
-	error("illegal file object%s%s\n",
-	      ((i->set_nonblocking_offset<0)?"; no set_nonblocking":""),
-	      ((i->set_blocking_offset<0)?"; no set_blocking":""));
+	 nobjects--;
+	 error("illegal file object%s%s\n",
+	       ((i->set_nonblocking_offset<0)?"; no set_nonblocking":""),
+	       ((i->set_blocking_offset<0)?"; no set_blocking":""));
       } else {
-        /* Try blocking mode */
-	i->type = I_BLOCKING_OBJ;
-	if (i==THIS->firstinput) {
-	  read_some_data();
-	}
-	return;
+	 /* Try blocking mode */
+	 i->type = I_BLOCKING_OBJ;
+	 if (i==THIS->firstinput) {
+	   /*
+	    * FIXME: What if read_som_data() returns 0?
+	    */
+	   read_some_data();
+	 }
+	 return;
       }
    }
   
@@ -778,8 +816,7 @@ static void pipe_write(INT32 args)
   i=new_input();
   i->type=I_STRING;
   nstrings++;
-  i->u.str=sp[-args].u.string;
-  i->u.str->refs++;
+  add_ref(i->u.str=sp[-args].u.string);
   pop_n_elems(args-1);
 }
 
@@ -827,7 +864,7 @@ static void pipe_output(INT32 args)
       {
 	b=THIS->firstbuffer;
 	THIS->firstbuffer=b->next;
-	lseek(THIS->fd,THIS->pos,0);
+	lseek(THIS->fd, THIS->pos, SEEK_SET);
 	write(THIS->fd,b->s->str,b->s->len);
 	sbuffers-=b->s->len;
 	nbuffers--;
@@ -847,7 +884,7 @@ static void pipe_output(INT32 args)
   } 
 
   THIS->living_outputs++;
-  THISOBJ->refs++;		/* Weird */
+  add_ref(THISOBJ);		/* Weird */
 
   /* Allocate a new struct output */
   obj=clone_object(output_program,0);
@@ -857,8 +894,7 @@ static void pipe_output(INT32 args)
   noutputs++;
   o->obj=NULL;
 
-  o->obj=sp[-args].u.object;
-  o->obj->refs++;
+  add_ref(o->obj=sp[-args].u.object);
 
   o->write_offset=find_identifier("write",o->obj->prog);
   o->set_nonblocking_offset=find_identifier("set_nonblocking",o->obj->prog);
@@ -879,8 +915,7 @@ static void pipe_output(INT32 args)
      o->pos=0; */
   o->pos=THIS->pos;
 
-  push_object(obj);
-  obj->refs++;
+  ref_push_object(obj);
   apply(o->obj,"set_id",1);
   pop_stack();
 
@@ -1073,7 +1108,7 @@ void close_and_free_everything(struct object *thisobj,struct pipe *p)
    p->done=1;
 
    if (thisobj) 
-      thisobj->refs++; /* don't kill object during this */
+      add_ref(thisobj); /* don't kill object during this */
 
    while (p->firstbuffer)
    {
