@@ -1,9 +1,11 @@
-/*\
+/*\ 
 ||| This file a part of Pike, and is copyright by Fredrik Hubinette
 ||| Pike is distributed as GPL (General Public License)
 ||| See the files COPYING and DISCLAIMER for more information.
 \*/
+/**/
 #include "global.h"
+#include "fdlib.h"
 #include "interpret.h"
 #include "svalue.h"
 #include "stralloc.h"
@@ -21,6 +23,8 @@
 #include "file_machine.h"
 #include "file.h"
 
+RCSID("$Id: efuns.c,v 1.65 1999/03/23 10:04:43 js Exp $");
+
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -31,25 +35,47 @@
 #include <signal.h>
 #include <errno.h>
 
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#endif /* HAVE_LIMITS_H */
+
 #if HAVE_DIRENT_H
 # include <dirent.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
-# define dirent direct
-# define NAMLEN(dirent) (dirent)->d_namlen
 # if HAVE_SYS_NDIR_H
 #  include <sys/ndir.h>
+#   define dirent direct
+#   define NAMLEN(dirent) (dirent)->d_namlen
 # endif
 # if HAVE_SYS_DIR_H
 #  include <sys/dir.h>
+#  define dirent direct
+#  define NAMLEN(dirent) (dirent)->d_namlen
 # endif
 # if HAVE_NDIR_H
 #  include <ndir.h>
+#  define dirent direct
+#  define NAMLEN(dirent) (dirent)->d_namlen
+# endif
+#endif
+
+#ifndef NAMELEN
+# if HAVE_DIRECT_H
+#  include <direct.h>
+#  define NAMLEN(dirent) strlen((dirent)->d_name)
 # endif
 #endif
 
 #include "dmalloc.h"
 
+#ifdef HAVE_PROCESS_H
+#include <process.h>
+#endif
+
+/* #define DEBUG_FILE */
+/* #define READDIR_DEBUG */
+
 struct array *encode_stat(struct stat *s)
 {
   struct array *a;
@@ -59,8 +85,15 @@ struct array *encode_stat(struct stat *s)
   {
   case S_IFREG: ITEM(a)[1].u.integer=s->st_size; break;
   case S_IFDIR: ITEM(a)[1].u.integer=-2; break;
+#ifdef S_IFLNK
   case S_IFLNK: ITEM(a)[1].u.integer=-3; break;
-  default: ITEM(a)[1].u.integer=-4; break;
+#endif
+  default:
+#ifdef DEBUG_FILE
+    fprintf(stderr, "encode_stat(): mode:%ld\n", (long)S_IFMT & s->st_mode);
+#endif /* DEBUG_FILE */
+    ITEM(a)[1].u.integer=-4;
+    break;
   }
   ITEM(a)[2].u.integer=s->st_atime;
   ITEM(a)[3].u.integer=s->st_mtime;
@@ -83,15 +116,25 @@ void f_file_stat(INT32 args)
     error("Bad argument 1 to file_stat()\n");
 
   s = sp[-args].u.string->str;
+
+#ifdef __NT__
+  /* Remove trailing / or \ on NT */
+  s = malloc(strlen(s)+1);
+  strcpy(s,sp[-args].u.string->str);
+  if(strlen(s) && (s[strlen(s)-1]=='/' || s[strlen(s)-1]=='\\'))
+    s[strlen(s)-1]=0;
+#endif
+  
   l = (args>1 && !IS_ZERO(sp+1-args))?1:0;
-  THREADS_ALLOW();
+  THREADS_ALLOW_UID();
+#ifdef HAVE_LSTAT
   if(l)
-  {
     i=lstat(s, &st);
-  }else{
+  else
+#endif
     i=stat(s, &st);
-  }
-  THREADS_DISALLOW();
+
+  THREADS_DISALLOW_UID();
   pop_n_elems(args);
   if(i==-1)
   {
@@ -99,8 +142,64 @@ void f_file_stat(INT32 args)
   }else{
     push_array(encode_stat(&st));
   }
+#ifdef __NT__
+  free(s);
+#endif  
 }
 
+#ifdef __NT__
+
+void f_filesystem_stat( INT32 args )
+{
+  char *path;
+  DWORD sectors_per_cluster = -1;
+  DWORD bytes_per_sector = -1;
+  DWORD free_clusters = -1;
+  DWORD total_clusters = -1;
+  char _p[4];
+  char *p = _p;
+  unsigned int free_sectors;
+  unsigned int total_sectors;
+
+  get_all_args( "filesystem_stat", args, "%s", &path );
+
+  if(sp[-1].u.string->len < 2 || path[1] != ':')
+  {
+    p = 0;
+  } else {
+    p[0] = path[0];
+    p[1] = ':';
+    p[2] = '\\';
+    p[3] = 0;
+  }
+  
+  if(!GetDiskFreeSpace( p, &sectors_per_cluster, 
+			&bytes_per_sector,
+			&free_clusters, 
+			&total_clusters ))
+  {
+    pop_n_elems(args);
+    push_int( 0 );
+    return;
+  }
+
+  free_sectors = sectors_per_cluster  * free_clusters;
+  total_sectors = sectors_per_cluster * total_clusters;
+  
+  pop_n_elems( args );
+  push_text("blocksize");
+  push_int(bytes_per_sector);
+  push_text("blocks");
+  push_int(total_sectors);
+  push_text("bfree");
+  push_int(free_sectors);
+  push_text("bavail");
+  push_int(free_sectors);
+  f_aggregate_mapping( 8 );
+}
+
+#else /* !__NT__ */
+
 #if  !defined(HAVE_STRUCT_STATFS) && !defined(HAVE_STRUCT_FS_DATA)
 #undef HAVE_STATFS
 #endif
@@ -209,15 +308,17 @@ void f_filesystem_stat(INT32 args)
     push_int(st.f_ffree);
     push_text("favail");
     push_int(st.f_favail);
+#if 0
     push_text("fsname");
     push_text(st.f_fstr);
+#endif /* 0 */
 #ifdef HAVE_STATVFS_F_BASETYPE
     push_text("fstype");
     push_text(st.f_basetype);
-    f_aggregate_mapping(9*2);
-#else /* !HAVE_STATVFS_ST_BASETYPE */
     f_aggregate_mapping(8*2);
-#endif /* HAVE_STATVFS_ST_BASETYPE */
+#else /* !HAVE_STATVFS_F_BASETYPE */
+    f_aggregate_mapping(7*2);
+#endif /* HAVE_STATVFS_F_BASETYPE */
 #else /* !HAVE_STATVFS */
 #ifdef HAVE_STATFS
 #ifdef HAVE_STRUCT_STATFS
@@ -276,13 +377,21 @@ void f_filesystem_stat(INT32 args)
 }
   
 #endif /* HAVE_STATVFS || HAVE_STATFS || HAVE_USTAT */
+#endif /* __NT__ */
 
 void f_werror(INT32 args)
 {
   if(!args)
-    error("Too few arguments to perror.\n");
+    error("Too few arguments to werror.\n");
   if(sp[-args].type != T_STRING)
-    error("Bad argument 1 to perror().\n");
+    error("Bad argument 1 to werror().\n");
+
+  if(args> 1)
+  {
+    extern void f_sprintf(INT32);
+    f_sprintf(args);
+    args=1;
+  }
 
   write_to_stderr(sp[-args].u.string->str, sp[-args].u.string->len);
   pop_n_elems(args);
@@ -302,8 +411,12 @@ void f_rm(INT32 args)
 
   s = sp[-args].u.string->str;
   
-  THREADS_ALLOW();
+  THREADS_ALLOW_UID();
+#ifdef HAVE_LSTAT
   i=lstat(s, &st) != -1;
+#else
+  i=stat(s, &st) != -1;
+#endif
   if(i)
   {
     if(S_IFDIR == (S_IFMT & st.st_mode))
@@ -313,7 +426,7 @@ void f_rm(INT32 args)
       i=unlink(s) != -1;
     }
   }
-  THREADS_DISALLOW();
+  THREADS_DISALLOW_UID();
       
   pop_n_elems(args);
   push_int(i);
@@ -322,33 +435,84 @@ void f_rm(INT32 args)
 void f_mkdir(INT32 args)
 {
   char *s;
-  INT32 i;
+  int mode;
+  int i;
+
   if(!args)
     error("Too few arguments to mkdir()\n");
 
   if(sp[-args].type != T_STRING)
     error("Bad argument 1 to mkdir().\n");
 
-  i=0770;
+  mode = 0770;
 
   if(args > 1)
   {
     if(sp[1-args].type != T_INT)
       error("Bad argument 2 to mkdir.\n");
 
-    i=sp[1-args].u.integer;
+    mode = sp[1-args].u.integer;
   }
   s=sp[-args].u.string->str;
-  THREADS_ALLOW();
-  i=mkdir(s, i) != -1;
-  THREADS_DISALLOW();
+#if MKDIR_ARGS == 2
+  THREADS_ALLOW_UID();
+  i = mkdir(s, mode) != -1;
+  THREADS_DISALLOW_UID();
+#else
+
+#ifdef HAVE_LSTAT
+#define LSTAT lstat
+#else
+#define LSTAT stat
+#endif
+
+  /* Most OS's should have MKDIR_ARGS == 2 nowadays fortunately. */
+  i = mkdir(s) != -1;
+
+  if (i) {
+    /* Attempt to set the mode.
+     *
+     * This code needs to be as paranoid as possible.
+     */
+    struct stat statbuf1;
+    struct stat statbuf2;
+    int mask = umask(0);
+    umask(mask);
+    i = LSTAT(s, &statbuf1) != -1;
+    if (i) {
+      i = ((statbuf1.st_mode & S_IFMT) == S_IFDIR);
+    }
+    if (i) {
+      mode = ((mode & 0777) | (statbuf1.st_mode & ~0777)) & ~mask;
+      i = chmod(s, mode) != -1;
+    }
+    if (i) {
+      i = LSTAT(s, &statbuf2) != -1;
+    }
+    if (i) {
+      i = (statbuf2.st_mode == mode) && (statbuf1.st_ino == statbuf2.st_ino);
+      if (!i) {
+	errno = EPERM;
+      }
+    }
+    /* The above code doesn't seem to work on NT, preliminary 'fix' follows. /js */
+#ifndef __NT__
+    if (!i) {
+      rmdir(s);
+    }
+#else
+    i=1;
+#endif
+    
+  }  
+#endif
   pop_n_elems(args);
   push_int(i);
 }
 
 #undef HAVE_READDIR_R
-#if defined(HAVE_SOLARIS_READDIR_R) || defined(HAVE_SOLARIS_HPUX_R) || \
-    defined(HAVE_POSIX_R)
+#if defined(HAVE_SOLARIS_READDIR_R) || defined(HAVE_HPUX_READDIR_R) || \
+    defined(HAVE_POSIX_READDIR_R)
 #define HAVE_READDIR_R
 #endif
 
@@ -363,9 +527,9 @@ void f_get_dir(INT32 args)
   get_all_args("get_dir",args,"%s",&path);
 
 #if defined(_REENTRANT) && defined(HAVE_READDIR_R)
-  THREADS_ALLOW();
+  THREADS_ALLOW_UID();
   dir=opendir(path);
-  THREADS_DISALLOW();
+  THREADS_DISALLOW_UID();
   if(dir)
   {
 #define FPR 1024
@@ -375,12 +539,15 @@ void f_get_dir(INT32 args)
     struct dirent *tmp;
 
     if (!(tmp =
-#ifdef HAVE_SOLARIS_READDIR_R
-	  alloca(sizeof(struct dirent) + 
+#if defined(HAVE_SOLARIS_READDIR_R) || defined(_PC_NAME_MAX)
+	  malloc(sizeof(struct dirent) + 
 		 ((pathconf(path, _PC_NAME_MAX) < 1024)?1024:
 		  pathconf(path, _PC_NAME_MAX)) + 1)
 #else
-	  alloca(sizeof(struct dirent) + NAME_MAX + 1024 + 1)
+#ifndef NAME_MAX
+#define NAME_MAX 1024
+#endif
+	  malloc(sizeof(struct dirent) + NAME_MAX+ 1024 + 1)
 #endif /* HAVE_SOLARIS_READDIR_R */
       )) {
       closedir(dir);
@@ -398,7 +565,6 @@ void f_get_dir(INT32 args)
 
       while(1)
       {
-	/* Should have code for the POSIX variant here also */
 #ifdef HAVE_SOLARIS_READDIR_R
 	/* Solaris readdir_r returns the second arg on success,
 	 * and returns NULL on error or at end of dir.
@@ -410,7 +576,11 @@ void f_get_dir(INT32 args)
 	if (!d) {
 	  /* Solaris readdir_r seems to set errno to ENOENT sometimes.
 	   */
-	  err = (errno == ENOENT)?0:errno;
+	  if (errno == ENOENT) {
+	    err = 0;
+	  } else {
+	    err = errno;
+	  }
 	  break;
 	}
 #elif defined(HAVE_HPUX_READDIR_R)
@@ -431,9 +601,36 @@ void f_get_dir(INT32 args)
 	/* POSIX readdir_r returns 0 on success, and ERRNO on failure.
 	 * at end of dir it sets the third arg to NULL.
 	 */
+	d = NULL;
+	errno = 0;
 	if ((err = readdir_r(dir, tmp, &d)) || !d) {
+#ifdef READDIR_DEBUG
+	  fprintf(stderr, "POSIX readdir_r(\"%s\") => err %d\n",
+		  path, err);
+	  fprintf(stderr, "POSIX readdir_r(), d= 0x%08x\n",
+		  (unsigned int)d);
+#endif /* READDIR_DEBUG */
+	  if (err == -1) {
+	    /* Solaris readdir_r returns -1, and sets errno. */
+	    err = errno;
+	  }
+#ifdef READDIR_DEBUG
+	  fprintf(stderr, "POSIX readdir_r(\"%s\") => errno %d\n",
+		  path, err);
+#endif /* READDIR_DEBUG */
+	  /* Solaris readdir_r seems to set errno to ENOENT sometimes.
+	   *
+	   * AIX readdir_r seems to set errno to EBADF at end of dir.
+	   */
+	  if ((err == ENOENT) || (err == EBADF)) {
+	    err = 0;
+	  }
 	  break;
 	}
+#ifdef READDIR_DEBUG
+	fprintf(stderr, "POSIX readdir_r(\"%s\") => \"%s\"\n",
+		path, d->d_name);
+#endif /* READDIR_DEBUG */
 #else
 #error Unknown readdir_r variant
 #endif
@@ -452,7 +649,7 @@ void f_get_dir(INT32 args)
       }
       THREADS_DISALLOW();
       if ((!d) && err) {
-	error("get_dir(): readdir_r() failed: %d\n", err);
+	error("get_dir(): readdir_r(\"%s\") failed: %d\n", path, err);
       }
       for(e=0;e<num_files;e++)
       {
@@ -464,6 +661,7 @@ void f_get_dir(INT32 args)
 	break;
     }
     THREADS_ALLOW();
+    free(tmp);
     closedir(dir);
     THREADS_DISALLOW();
     a=aggregate_array(sp-save_sp);
@@ -529,9 +727,9 @@ void f_getcwd(INT32 args)
 #define MAXPATHLEN 32768
 #endif
   tmp=xalloc(MAXPATHLEN+1);
-  THREADS_ALLOW();
+  THREADS_ALLOW_UID();
   e=(char *)getwd(tmp);
-  THREADS_DISALLOW();
+  THREADS_DISALLOW_UID();
 #endif
   if(!e) {
     if (tmp) 
@@ -544,38 +742,15 @@ void f_getcwd(INT32 args)
   free(e);
 }
 
-void f_fork(INT32 args)
-{
-  int res;
-  do_set_close_on_exec();
-  pop_n_elems(args);
-#if defined(HAVE_FORK1) && defined(_REENTRANT)
-  res = fork1();
-#else
-  res = fork();
-#endif
-  push_int(res);
-  if(!res && res!=-1)
-  {
-    call_callback(&fork_child_callback, 0);
-  }
-#if defined(_REENTRANT)
-  if(!res)
-  {
-    /* forked copy. there is now only one thread running, this one. */
-    num_threads=1;
-  }
-#endif
-}
-
-
 void f_exece(INT32 args)
 {
   INT32 e;
   char **argv, **env;
-  extern char **environ;
   struct svalue *save_sp;
   struct mapping *en;
+#ifdef DECLARE_ENVIRON
+  extern char **environ;
+#endif
 
   save_sp=sp-args;
 
@@ -654,7 +829,18 @@ void f_exece(INT32 args)
   my_set_close_on_exec(2,0);
 
   do_set_close_on_exec();
-  execve(argv[0],argv,env);
+
+#ifdef __NT__
+#define DOCAST(X) ((const char * const *)(X))
+#else
+#define DOCAST(X) (X)
+#endif
+
+#ifdef HAVE_BROKEN_F_SETFD
+  do_close_on_exec();
+#endif /* HAVE_BROKEN_F_SETFD */
+
+  execve(argv[0],DOCAST(argv),DOCAST(env));
 
   free((char *)argv);
   if(env != environ) free((char *)env);
@@ -717,19 +903,18 @@ void init_files_efuns(void)
 
   add_efun("file_stat",f_file_stat,
 	   "function(string,int|void:int *)", OPT_EXTERNAL_DEPEND);
-#if defined(HAVE_STATVFS) || defined(HAVE_STATFS) || defined(HAVE_USTAT)
+#if defined(HAVE_STATVFS) || defined(HAVE_STATFS) || defined(HAVE_USTAT) || defined(__NT__)
   add_efun("filesystem_stat", f_filesystem_stat,
 	   "function(string:mapping(string:string|int))", OPT_EXTERNAL_DEPEND);
 #endif /* HAVE_STATVFS || HAVE_STATFS */
   add_efun("errno",f_errno,"function(:int)",OPT_EXTERNAL_DEPEND);
-  add_efun("werror",f_werror,"function(string:void)",OPT_SIDE_EFFECT);
+  add_efun("werror",f_werror,"function(string,void|mixed...:void)",OPT_SIDE_EFFECT);
   add_efun("rm",f_rm,"function(string:int)",OPT_SIDE_EFFECT);
   add_efun("mkdir",f_mkdir,"function(string,void|int:int)",OPT_SIDE_EFFECT);
   add_efun("mv", f_mv, "function(string,string:int)", OPT_SIDE_EFFECT);
   add_efun("get_dir",f_get_dir,"function(string:string *)",OPT_EXTERNAL_DEPEND);
   add_efun("cd",f_cd,"function(string:int)",OPT_SIDE_EFFECT);
   add_efun("getcwd",f_getcwd,"function(:string)",OPT_EXTERNAL_DEPEND);
-  add_efun("fork",f_fork,"function(:int)",OPT_SIDE_EFFECT);
   add_efun("exece",f_exece,"function(string,mixed*,void|mapping(string:string):int)",OPT_SIDE_EFFECT); 
 
 #ifdef HAVE_STRERROR
