@@ -27,6 +27,7 @@
 #include <unistd.h>
 #endif
 
+#include "fdlib.h"
 #include "stralloc.h"
 #include "pike_macros.h"
 #include "machine.h"
@@ -42,7 +43,7 @@
 #include "threads.h"
 #include "operators.h"
 
-RCSID("$Id: spider.c,v 1.58 1998/07/16 18:22:38 grubba Exp $");
+RCSID("$Id: spider.c,v 1.77 1998/12/05 04:01:03 hubbe Exp $");
 
 #ifdef HAVE_PWD_H
 #include <pwd.h>
@@ -158,8 +159,7 @@ void f_parse_accessed_database(INT32 args)
 
   push_string(make_shared_string("\n"));
   f_divide(2);
-  arg = sp[-1].u.array;
-  arg->refs++;
+  add_ref(arg = sp[-1].u.array);
   /* The initial string is gone, but the array is there now. */
   pop_stack();
 
@@ -334,18 +334,15 @@ void f_parse_html(INT32 args)
     return;
   }
 
-  ss->refs++;
+  add_ref(ss);
 
-  single=sp[1-args].u.mapping; 
-  cont=sp[2-args].u.mapping; 
-  cont->refs++;
-  single->refs++;
+  add_ref(single=sp[1-args].u.mapping);
+  add_ref(cont=sp[2-args].u.mapping);
 
   if (args>3)
   {
     f_aggregate(args-3);
-    extra_args=sp[-1].u.array;
-    extra_args->refs++;
+    add_ref(extra_args=sp[-1].u.array);
     pop_stack();
   }
   else extra_args=NULL;
@@ -389,33 +386,31 @@ void f_parse_html_lines(INT32 args)
 
   sp[-args].type=T_INT;
 
-  single=sp[1-args].u.mapping; 
-  cont=sp[2-args].u.mapping; 
-  cont->refs++;
-  single->refs++;
-
+  add_ref(single=sp[1-args].u.mapping);
+  add_ref(cont=sp[2-args].u.mapping);
+ 
   if (args>3)
   {
     f_aggregate(args-3);
-    extra_args=sp[-1].u.array;
-    extra_args->refs++;
+    add_ref(extra_args=sp[-1].u.array);
     pop_stack();
   }
   else extra_args=NULL;
 
   pop_n_elems(3);
+/*   fprintf(stderr, "sp=%p\n", sp); */
 
   strings=0;
   do_html_parse_lines(ss,cont,single,&strings,MAX_PARSE_RECURSE,extra_args,1);
 
   if (extra_args) free_array(extra_args);
-
   free_mapping(cont);
   free_mapping(single);
   if(strings > 1)
     f_add(strings);
   else if(!strings)
     push_text("");
+/*   fprintf(stderr, "sp=%p (strings=%d)\n", sp, strings); */
 }
 
 char start_quote_character = '\000';
@@ -657,8 +652,7 @@ void do_html_parse(struct pike_string *ss,
 
       push_string(make_shared_binary_string((char *)s+n, j-n));
       f_lower_case(1);
-      sval2.u.string = sp[-1].u.string;
-      sval2.u.string->refs++;
+      add_ref(sval2.u.string = sp[-1].u.string);
       sval2.type=T_STRING;
       pop_stack();
 
@@ -667,13 +661,34 @@ void do_html_parse(struct pike_string *ss,
 
       if (sval1.type==T_STRING)
       {
+	int quote = 0;
 	/* A simple string ... */
+	if (last < i-1)
+	{ 
+	  push_string(make_shared_binary_string(s+last, i-last-1)); 
+	  (*strings)++; 
+	}
+
 	assign_svalue_no_free(sp++,&sval1);
-	free_svalue(&sval1);
 	(*strings)++;
-	find_endtag(sval2.u.string ,s+j, len-j, &l);
+	free_svalue(&sval1);
 	free_svalue(&sval2);
-	j+=l;
+
+	/* Scan ahead to the end of the tag... */
+	for (; j<len; j++) {
+	  if (quote) {
+	    if (s[j] == quote) {
+	      quote = 0;
+	    }
+	  } else if (s[j] == '>') {
+	    break;
+	  } else if ((s[j] == '\'') || (s[j] == '\"')) {
+	    quote = s[j];
+	  }
+	}
+	if (j < len) {
+	  j++;
+	}
 	i=last=j;
 	continue;
       }
@@ -681,10 +696,10 @@ void do_html_parse(struct pike_string *ss,
       {
 	/* Hopefully something callable ... */
 	assign_svalue_no_free(sp++,&sval2);
-	k=push_parsed_tag(s+j,len-j); 
+	k = push_parsed_tag(s+j,len-j); 
 	if (extra_args)
 	{
-	  extra_args->refs++;
+	  add_ref(extra_args);
 	  push_array_items(extra_args);
 	}
 
@@ -730,9 +745,16 @@ void do_html_parse(struct pike_string *ss,
       mapping_index_no_free(&sval1,cont,&sval2);
       if (sval1.type==T_STRING)
       {
+	if (last < i-1)
+	{ 
+	  push_string(make_shared_binary_string(s+last, i-last-1)); 
+	  (*strings)++; 
+	}
+
 	assign_svalue_no_free(sp++,&sval1);
 	free_svalue(&sval1);
 	(*strings)++;
+
 	find_endtag(sval2.u.string,s+j,len-j,&l);
 	free_svalue(&sval2);
 	j+=l;
@@ -741,16 +763,16 @@ void do_html_parse(struct pike_string *ss,
       }
       else if (sval1.type != T_INT)
       {
-	assign_svalue_no_free(sp++,&sval2);
-	m=push_parsed_tag(s+j,len-j) + j;
-	k=find_endtag(sval2.u.string,s+m,len-m,&l);
-	push_string(make_shared_binary_string(s+m,k));
-	m+=l;
+	assign_svalue_no_free(sp++, &sval2);
+	m = push_parsed_tag(s+j, len-j) + j;
+	k = find_endtag(sval2.u.string, s+m, len-m, &l);
+	push_string(make_shared_binary_string(s+m, k));
+	m += l;
         /* M == just after end tag, from s */
 
 	if (extra_args)
 	{
-	  extra_args->refs++;
+	  add_ref(extra_args);
 	  push_array_items(extra_args);
 	}
 
@@ -763,11 +785,11 @@ void do_html_parse(struct pike_string *ss,
 	  copy_shared_string(ss2,sp[-1].u.string);
 	  pop_stack();
 
-	  /* i == first '<' ? */
-	  /* last == end of last tag */
-	  if (last!=i-1)
+	  /* i == '<' + 1 */
+	  /* last == end of previous tags '>' + 1 */
+	  if (last < i-1)
 	  { 
-	    push_string(make_shared_binary_string(s+last,i-last-1)); 
+	    push_string(make_shared_binary_string(s+last, i-last-1)); 
 	    (*strings)++; 
 	  }
 	  i=last=j=m;
@@ -780,9 +802,9 @@ void do_html_parse(struct pike_string *ss,
 	  copy_shared_string(ss2,sp[-1].u.string);
 	  pop_stack();
 
-	  if (last!=i-1)
+	  if (last < i-1)
 	  { 
-	    push_string(make_shared_binary_string(s+last,i-last-1)); 
+	    push_string(make_shared_binary_string(s+last, i-last-1)); 
 	    (*strings)++; 
 	  }
 	  i=last=j=m;
@@ -874,6 +896,9 @@ void do_html_parse_lines(struct pike_string *ss,
   char *s;
   struct svalue sval1,sval2;
   struct pike_string *ss2;
+
+/*   fprintf(stderr, "sp=%p (strings=%d)\n", sp, *strings); */
+
   if (!ss->len)
   {
     free_string(ss);
@@ -891,6 +916,8 @@ void do_html_parse_lines(struct pike_string *ss,
   len=ss->len;
 
   last=0;
+
+
   for (i=0; i<len-1;)
   {
     if (s[i]==10)
@@ -906,56 +933,78 @@ void do_html_parse_lines(struct pike_string *ss,
 
       push_string(make_shared_binary_string((char *)s+i, j-i));
       f_lower_case(1);
-      sval2.u.string = sp[-1].u.string;
-      sval2.u.string->refs++;
+      add_ref(sval2.u.string = sp[-1].u.string);
       sval2.type=T_STRING;
       pop_stack();
 
       /* Is this a non-container? */
       mapping_index_no_free(&sval1,single,&sval2);
-
+/*       if(sval1.type == T_INT) */
+/* 	mapping_index_no_free(&sval1,single,&empty_string); */
+	
       if (sval1.type==T_STRING)
       {
+	int quote = 0;
 	/* A simple string ... */
+	if (last < i-1)
+	{ 
+	  push_string(make_shared_binary_string(s+last, i-last-1)); 
+	  (*strings)++; 
+	}
+
 	*(sp++)=sval1;
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
 	sval1.type=99;
 #endif
 	(*strings)++;
-	/* find_endtag(sval2.u.string ,s+j, len-j, &l);  bug /law 960805 */
 	free_svalue(&sval2);
-	j+=l;
+
+	/* Scan ahead to the end of the tag... */
+	for (; j<len; j++) {
+	  if (quote) {
+	    if (s[j] == quote) {
+	      quote = 0;
+	    }
+	  } else if (s[j] == '>') {
+	    break;
+	  } else if ((s[j] == '\'') || (s[j] == '\"')) {
+	    quote = s[j];
+	  }
+	}
+	if (j < len) {
+	  j++;
+	}
 	i=last=j;
 	continue;
       }
       else if (sval1.type!=T_INT)
       {
 	*(sp++)=sval2;
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
 	sval2.type=99;
 #endif
 	k=push_parsed_tag(s+j,len-j);
 	push_int(line);
 	if (extra_args)
 	{
-	  extra_args->refs++;
+	  add_ref(extra_args);
 	  push_array_items(extra_args);
 	}
 	apply_svalue(&sval1,3+(extra_args?extra_args->size:0));
 	HANDLE_RETURN_VALUE(j+k);
 	continue;
       }
-      /* free_svalue(&sval1); Not needed. The type is always T_INT */
 
+      /* free_svalue(&sval1); Not needed. The type is always T_INT */
       /* Is it a container then? */
+
       mapping_index_no_free(&sval1,cont,&sval2);
-      if (sval1.type == T_INT) {
+      if(sval1.type == T_INT)
 	mapping_index_no_free(&sval1,cont,&empty_string);
-      }
       if (sval1.type==T_STRING)
       {
 	*(sp++)=sval1;
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
 	sval1.type=99;
 #endif
 	(*strings)++;
@@ -969,18 +1018,18 @@ void do_html_parse_lines(struct pike_string *ss,
       else if (sval1.type != T_INT)
       {
 	*(sp++)=sval2;
-#ifdef DEBUG
+#ifdef PIKE_DEBUG
 	sval2.type=99;
 #endif
-	m=push_parsed_tag(s+j,len-j) + j;
-	k=find_endtag(sval2.u.string,s+m,len-m,&l);
-	push_string(make_shared_binary_string(s+m,k));
-	m+=l;
+	m = push_parsed_tag(s+j, len-j) + j;
+	k = find_endtag(sval2.u.string, s+m, len-m, &l);
+	push_string(make_shared_binary_string(s+m, k));
+	m += l;
         /* M == just after end tag, from s */
 	push_int(line);
 	if (extra_args)
 	{
-	  extra_args->refs++;
+	  add_ref(extra_args);
 	  push_array_items(extra_args);
 	}
 	apply_svalue(&sval1,4+(extra_args?extra_args->size:0));
@@ -1002,7 +1051,7 @@ void do_html_parse_lines(struct pike_string *ss,
   }
   else if (last<len)
   {
-    push_string(make_shared_binary_string(s+last,len-last));  
+    push_string(make_shared_binary_string(s+last,len-last));
     free_string(ss);
     (*strings)++;
   }
@@ -1132,8 +1181,7 @@ void f_mark_fd(INT32 args)
     {
       if(fd_marks[fd])
       {
-	fd_marks[fd]->refs++;
-	push_string(fd_marks[fd]);
+	ref_push_string(fd_marks[fd]);
       } else {
 	push_text("");
       }
@@ -1149,8 +1197,7 @@ void f_mark_fd(INT32 args)
     }
   }
   
-  s=sp[-args+1].u.string;
-  s->refs++;
+  add_ref(s=sp[-args+1].u.string);
   if(fd_marks[fd])
     free_string(fd_marks[fd]);
   fd_marks[fd]=s;
@@ -1189,8 +1236,7 @@ void f_fcgi_create_listen_socket(INT32 args)
 static void program_name(struct program *p)
 {
   char *f;
-  p->refs++;
-  push_program(p);
+  ref_push_program(p);
   APPLY_MASTER("program_name", 1);
   if(sp[-1].type == T_STRING)
     return;
@@ -1253,11 +1299,12 @@ static struct program *streamed_parser;
 #endif /* ENABLE_STREAMED_PARSER */
 
 extern void init_udp(void);
+extern void init_xml(void);
 
 
 /* Hohum. Here we go. This is try number three for a more optimized Roxen. */
 
-#if 0 && defined(_REENTRANT)
+#ifdef _REENTRANT
 #define BUFFER (8192)
 
 struct thread_args
@@ -1271,47 +1318,49 @@ struct thread_args
   int len;
   int sent;
   char buffer[BUFFER];
-  THREAD_T tid;
 };
 
 MUTEX_T done_lock;
 struct thread_args *done;
 
 /* WARNING! This function is running _without_ any stack etc. */
-void *do_shuffle(void *_a)
+
+#define MY_MIN(a,b) ((a)<(b)?(a):(b))
+void do_shuffle(void *_a)
 {
   struct thread_args *a = (struct thread_args *)_a;
 
 #ifdef DIRECTIO_ON
-  if(a->len >= 65536*2)
+  if(a->len > (65536*2))
     directio(a->from_fd, DIRECTIO_ON);
 #endif
 
   while(a->len)
   {
-    int nread;
-    int start = 0;
-
-    nread = read(a->from_fd, a->buffer, BUFFER);
+    int nread, written=0;
+    nread = fd_read(a->from_fd, a->buffer, MY_MIN(BUFFER,a->len));
     if(nread <= 0) {
-      if (errno != EINTR)
+      if (!nread)
 	break;
-      else
+      if(errno == EINTR)
 	continue;
+      else
+	break;
     }
 
     while(nread)
     {
-      int nsent = write(a->to_fd, a->buffer+start, nread);
+      int nsent = fd_write(a->to_fd, a->buffer+written, nread);
       if(nsent < 0) {
-	if (errno != EINTR)
+	if(errno != EINTR)
 	  goto end;
-      } else {
-	start += nsent;
-	a->sent += nsent;
-	nread -= nsent;
-	a->len -= nsent;
+	else 
+	  continue;
       }
+      written += nsent;
+      a->sent += nsent;
+      nread -= nsent;
+      a->len -= nsent;
     }
   }
 
@@ -1324,7 +1373,6 @@ void *do_shuffle(void *_a)
   done = a;
   mt_unlock(&done_lock);
   wake_up_backend();
-  return 0;
 }
 
 static int num_shuffles = 0;
@@ -1372,13 +1420,13 @@ void f_shuffle(INT32 args)
   apply(a->from, "query_fd", 0);
   get_all_args("shuffle", 2, "%d%d", &a->to_fd, &a->from_fd);
 
-  a->from->refs++; 
-  a->to->refs++; 
+  add_ref(a->from);
+  add_ref(a->to); 
   
   assign_svalue_no_free(&a->cb, q);
   assign_svalue_no_free(&a->args, w);
   
-  th_create_small(&a->tid, do_shuffle, (void *)a);
+  th_farm(do_shuffle, (void *)a);
 
   if(!my_callback)
     my_callback = add_backend_callback( finished_p, 0, 0 );
@@ -1390,12 +1438,12 @@ void f_shuffle(INT32 args)
 
 void pike_module_init(void) 
 {
-  push_text("");
+  ref_push_string(make_shared_string(""));
   empty_string = sp[-1];
-  empty_string.u.string->refs++;
   pop_stack();
 
-#if 0 && defined(_REENTRANT)
+
+#ifdef _REENTRANT
   add_function("shuffle", f_shuffle, 
 	       "function(object,object,function,mixed,int:void)", 0);
 #endif
@@ -1424,11 +1472,11 @@ void pike_module_init(void)
 	   OPT_EXTERNAL_DEPEND);
 
   add_efun("parse_html",f_parse_html,
-	   "function(string,mapping(string:function(string|void,mapping(string:string)|void,mixed ...:string)),mapping(string:function(string|void,mapping(string:string)|void,string|void,mixed ...:string)),mixed ...:string)",
+	   "function(string,mapping(string:string|function(string|void,mapping(string:string)|void,mixed ...:string)),mapping(string:string|function(string|void,mapping(string:string)|void,string|void,mixed ...:string)),mixed ...:string)",
 	   0);
 
   add_efun("parse_html_lines",f_parse_html_lines,
-	   "function(string,mapping(string:function(string|void,mapping(string:string)|void,int|void,mixed ...:string)),mapping(string:function(string|void,mapping(string:string)|void,string|void,int|void,mixed ...:string)),mixed ...:string)",
+	   "function(string,mapping(string:string|function(string|void,mapping(string:string)|void,int|void,mixed ...:string)),mapping(string:string|function(string|void,mapping(string:string)|void,string|void,int|void,mixed ...:string)),mixed ...:string)",
 	   0);
   
 #ifdef HAVE_PERROR
@@ -1478,6 +1526,8 @@ void pike_module_init(void)
   streamed_parser = end_program();
   add_program_constant("streamed_parser", streamed_parser,0);
 #endif /* ENABLE_STREAMED_PARSER */
+
+  init_xml();
 }
 
 
